@@ -2,6 +2,25 @@
 
 #include "Python.h"
 
+#if PY_MAJOR_VERSION == 2
+#define PyLong_AsLong PyInt_AsLong
+#define PyLong_FromLong PyInt_FromLong
+#define PyUnicode_Check3 PyString_Check
+#define PyUnicode_FromFormat PyString_FromFormat
+#define PyUnicode_FromString PyString_FromString
+#define PyUnicode_FromStringAndSize PyString_FromStringAndSize
+#define PyUnicode_Type PyString_Type
+#define PyUnicode_AsUTF8 PyString_AsString
+#define PyVarObject_HEAD_INIT(p, b) PyObject_HEAD_INIT(p) 0,
+#define OB_TYPE ob_type
+#else
+#define PyUnicode_Check3 PyUnicode_Check
+#define OB_TYPE ob_base.ob_type
+#if PY_MINOR_VERSION == 2
+#define PyUnicode_AsUTF8 _PyUnicode_AsString
+#endif
+#endif
+
 #include "datetime.h"
 #include "structmember.h"
 
@@ -101,6 +120,12 @@ typedef struct {
      PyObject *catalog_number;	/* TLE catalog number */
 } EarthSatellite;
 
+/* A forward-reference function definition, so that C does not have to
+   do anything crazy like make two passes over its input file.  Keep it
+   cool, C, I will just repeat myself instead. */
+
+static int Body_obj_cir(Body *body, char *fieldname, unsigned topocentric);
+
 /* Return the mjd of the current time. */
 
 static double mjd_now(void)
@@ -119,38 +144,6 @@ static int PyNumber_AsDouble(PyObject *o, double *dp)
      return 0;
 }
 
-/* My own version of the PyString_AsString function that Python used to
-   have, with the main difference from the code's point of view being
-   that it now requires a free() of the memory it returns.  The argument
-   must be a PyUnicode object. */
-
-static char *MyString_AsString(PyObject *o)
-{
-     Py_ssize_t i;
-     Py_ssize_t size = PyUnicode_GET_SIZE(o);
-     Py_UNICODE *u = PyUnicode_AS_UNICODE(o);
-     char *s = malloc(size + 1);
-     if (!s) {
-          PyErr_SetString(PyExc_RuntimeError, "cannot malloc() storage"
-                          " for the ASCII version of your string");
-          return 0;
-     }
-     for (i=0; i < size; i++) {
-          if ((u[i] >= 32 && u[i] < 127)
-               || u[i] == '\n' || u[i] == '\t' || u[i] == '\r') {
-               s[i] = u[i];
-          } else {
-               PyErr_SetString(PyExc_ValueError, "strings passed to this"
-                               " routine must consist of printable ASCII"
-                               " characters only");
-               free(s);
-               return 0;
-          }
-     }
-     s[i] = '\0';
-     return s;
-}
-
 /* The libastro library offers a "getBuiltInObjs()" function that
    initializes the list of planets that XEphem displays by default.
    Rather than duplicate its logic for how to build each objects, we
@@ -292,6 +285,9 @@ static PyObject *Angle_get_znorm(PyObject *self, void *v)
 
 static PyNumberMethods Angle_NumberMethods = {
      NULL, NULL, NULL, NULL, NULL, NULL, /* skip six fields */
+#if PY_MAJOR_VERSION == 2
+     NULL,
+#endif
      Angle_neg, /* nb_negative */
      Angle_pos, /* nb_positive */
      NULL
@@ -379,7 +375,7 @@ static int parse_mjd_from_string(PyObject *so, double *mjdp)
 
      if (len >= 1) {
 	  int i;
-          char *s = MyString_AsString(PyList_GetItem(pieces, 0));
+          char *s = PyUnicode_AsUTF8(PyList_GetItem(pieces, 0));
           if (!s) goto fail;
 
 	  /* Make sure all characters are in set '-/.0123456789' */
@@ -387,24 +383,20 @@ static int parse_mjd_from_string(PyObject *so, double *mjdp)
 	  for (i=0; s[i]; i++) {
 	       if (s[i] != '-' && s[i] != '/' && s[i] != '.'
 		   && (s[i] < '0' || s[i] > '9')) {
-                    free(s);
 		    goto fail;
                }
           }
 
 	  f_sscandate(s, PREF_YMD, &month, &day, &year);
-          free(s);
      }
 
      if (len >= 2) {
-          char *t = MyString_AsString(PyList_GetItem(pieces, 1));
+          char *t = PyUnicode_AsUTF8(PyList_GetItem(pieces, 1));
 	  double hours;
           if (!t) goto fail;
 	  if (f_scansexa(t, &hours) == -1) {
-               free(t);
 	       goto fail;
           }
-          free(t);
 	  day += hours / 24.;
      }
 
@@ -417,7 +409,7 @@ fail:
      if (! PyErr_Occurred()) {
 	  PyObject *repr = PyObject_Repr(so);
           PyObject *complaint = PyUnicode_FromFormat(
-	       "your date string %u does not look like a year/month/day"
+	       "your date string %s does not look like a year/month/day"
 	       " optionally followed by hours:minutes:seconds", repr);
 	  PyErr_SetObject(PyExc_ValueError, complaint);
 	  Py_DECREF(repr);
@@ -450,6 +442,8 @@ static int parse_mjd_from_datetime(PyObject *value, double *mjdp)
           *mjdp += PyDateTime_DATE_GET_HOUR(value) / 24.;
           *mjdp += PyDateTime_DATE_GET_MINUTE(value) / (24. * 60.);
           *mjdp += PyDateTime_DATE_GET_SECOND(value) / (24. * 60. * 60.);
+          *mjdp += PyDateTime_DATE_GET_MICROSECOND(value)
+               / (24. * 60. * 60. * 1000000.);
      }
      return 0;
 }
@@ -458,7 +452,7 @@ static int parse_mjd(PyObject *value, double *mjdp)
 {
      if (PyNumber_Check(value))
 	  return parse_mjd_from_number(value, mjdp);
-     else if (PyUnicode_Check(value))
+     else if (PyUnicode_Check3(value))
 	  return parse_mjd_from_string(value, mjdp);
      else if (PyTuple_Check(value))
 	  return parse_mjd_from_tuple(value, mjdp);
@@ -509,7 +503,9 @@ static char *Date_format(PyObject *self)
      static char buffer[64];
      int year, month, day, hour, minute;
      double second;
-     mjd_six(d->ob_fval, &year, &month, &day, &hour, &minute, &second);
+     /* Note the offset, which makes us round instead of truncate. */
+     mjd_six(d->ob_fval + 0.5 / 24.0 / 60.0 / 60.0,
+             &year, &month, &day, &hour, &minute, &second);
      sprintf(buffer, "%d/%d/%d %02d:%02d:%02d",
 	     year, month, day, hour, minute, (int) second);
      return buffer;
@@ -637,12 +633,15 @@ static int parse_angle(PyObject *value, double factor, double *result)
 {
      if (PyNumber_Check(value)) {
 	  return PyNumber_AsDouble(value, result);
-     } else if (PyUnicode_Check(value)) {
+     } else if (PyUnicode_Check3(value)) {
 	  double scaled;
-	  char *s = MyString_AsString(value);
+	  char *s = PyUnicode_AsUTF8(value);
           if (!s) return -1;
-          f_scansexa(s, &scaled);
-          free(s);
+          if (f_scansexa(s, &scaled) == -1) {
+               PyErr_Format(PyExc_ValueError, "your angle string %r does not "
+                            "have the format [number[:number[:number]]]", s);
+               return -1;
+	  }
 	  *result = scaled / factor;
 	  return 0;
      } else {
@@ -655,6 +654,16 @@ static int parse_angle(PyObject *value, double factor, double *result)
 static double to_angle(PyObject *value, double efactor, int *status)
 {
      double r;
+#if PY_MAJOR_VERSION == 2
+     /* Support Unicode strings under Python 2 */
+     if (PyUnicode_Check(value)) {
+          value = PyUnicode_AsUTF8String(value);
+          if (!value) {
+	       *status = -1;
+               return 0;
+          }
+     }
+#endif
      if (PyNumber_Check(value)) {
 	  value = PyNumber_Float(value);
 	  if (!value) {
@@ -665,20 +674,22 @@ static double to_angle(PyObject *value, double efactor, int *status)
 	  Py_DECREF(value);
 	  *status = 0;
 	  return r;
-     } else if (PyUnicode_Check(value)) {
+     } else if (PyUnicode_Check3(value)) {
 	  double scaled;
-	  char *s = MyString_AsString(value);
+          char *s = PyUnicode_AsUTF8(value);
           if (!s) {
                *status = -1;
                return 0;
           }
-          f_scansexa(s, &scaled);
-          free(s);
-	  *status = 0;
+          *status = f_scansexa(s, &scaled);
+          if (*status == -1) {
+              PyErr_Format(PyExc_ValueError, "your angle string %r does not "
+                           "have the format [number[:number[:number]]]", s);
+          }
 	  return scaled / efactor;
      } else {
 	  PyErr_SetString(PyExc_TypeError,
-			  "can only update value with String or number");
+			  "can only update value with string or number");
 	  *status = -1;
 	  return 0;
      }
@@ -792,25 +803,21 @@ static PyObject* get_f_spect(PyObject *self, void *v)
 static int set_f_spect(PyObject *self, PyObject *value, void *v)
 {
      Body *b = (Body*) self;
-     Py_UNICODE *u;
-     if (!PyUnicode_Check(value)) {
-	  PyErr_SetString(PyExc_ValueError,
-                          "spectral code must be a string");
+     char *s;
+     if (!PyUnicode_Check3(value)) {
+	  PyErr_SetString(PyExc_ValueError, "spectral code must be a string");
 	  return -1;
      }
-     if (PyUnicode_GET_SIZE(value) != 2) {
+     s = PyUnicode_AsUTF8(value);
+     if (!s)
+          return -1;
+     if (s[0] == '\0' || s[1] == '\0' || s[2] != '\0') {
 	  PyErr_SetString(PyExc_ValueError,
 			  "spectral code must be two characters long");
 	  return -1;
      }
-     u = PyUnicode_AS_UNICODE(value);
-     if (u[0] < 32 || u[1] < 32 || u[0] > 126 || u[1] > 126) {
-	  PyErr_SetString(PyExc_ValueError,
-			  "spectral code must contain only ASCII characters");
-	  return -1;
-     }
-     b->obj.f_spect[0] = u[0];
-     b->obj.f_spect[1] = u[1];
+     b->obj.f_spect[0] = s[0];
+     b->obj.f_spect[1] = s[1];
      return 0;
 }
 
@@ -1280,6 +1287,23 @@ fail:
      return 0;
 }
 
+static PyObject* Body_parallactic_angle(PyObject *self)
+{
+     PyObject *a1, *a2;
+     Body *body = (Body*) self;
+     double ha, pa;
+     if (Body_obj_cir(body, "parallactic_angle", 1) == -1)
+          return 0;
+     radec2ha(&(body->now), body->obj.s_astrora, body->obj.s_astrodec, &ha);
+     pa = parallacticLHD(body->now.n_lat, ha, body->obj.s_astrodec);
+     a1 = new_Angle(pa, raddeg(1));
+     if (!a1)
+          return 0;
+     a2 = Angle_get_znorm(a1, 0);
+     Py_XDECREF(a1);
+     return a2;
+}
+
 static PyObject* Body_writedb(PyObject *self)
 {
      Body *body = (Body*) self;
@@ -1293,6 +1317,7 @@ static PyObject* Body_copy(PyObject *self)
      PyObject *newbody = _PyObject_New(self->ob_type);
      if (!newbody) return 0;
      memcpy(newbody, self, self->ob_type->tp_basicsize);
+     newbody->ob_refcnt = 1;  /* since memcpy will have overwritten it */
      Py_XINCREF(((Body*) self)->name);
      return newbody;
 }
@@ -1301,23 +1326,37 @@ static PyObject* Body_repr(PyObject *body_object)
 {
      Body *body = (Body*) body_object;
      if (body->name) {
-	  return PyUnicode_FromFormat("<%s %R at %p>",
-                                      body->ob_base.ob_type->tp_name,
-                                      body->name, body);
+	  char *name;
+	  PyObject *repr, *result;
+	  repr = PyObject_Repr(body->name);
+	  if (!repr)
+               return 0;
+	  name = PyUnicode_AsUTF8(repr);
+	  if (!name) {
+               Py_DECREF(repr);
+               return 0;
+          }
+	  result = PyUnicode_FromFormat("<%s %s at %p>",
+				       body->OB_TYPE->tp_name, name, body);
+	  Py_DECREF(repr);
+	  return result;
      } else if (body->obj.o_name[0])
 	  return PyUnicode_FromFormat("<%s \"%s\" at %p>",
-                                      body->ob_base.ob_type->tp_name,
-                                      body->obj.o_name, body);
+				     body->OB_TYPE->tp_name,
+				     body->obj.o_name, body);
      else
 	  return PyUnicode_FromFormat("<%s at %p>",
-                                      body->ob_base.ob_type->tp_name,
-                                      body);
+				     body->OB_TYPE->tp_name, body);
 }
 
 static PyMethodDef Body_methods[] = {
      {"compute", (PyCFunction) Body_compute, METH_VARARGS | METH_KEYWORDS,
       "compute the location of the body for the given date or Observer,"
       " or for the current time if no date is supplied"},
+     {"parallactic_angle", (PyCFunction) Body_parallactic_angle, METH_NOARGS,
+      "return the parallactic angle to the body; an Observer must have been"
+      " provided to the most recent compute() call, because a parallactic"
+      " angle is always measured with respect to a specfic observer"},
      {"writedb", (PyCFunction) Body_writedb, METH_NOARGS,
       "return a string representation of the body "
       "appropriate for inclusion in an ephem database file"},
@@ -1615,10 +1654,10 @@ static PyObject *Get_name(PyObject *self, void *v)
 static int Set_name(PyObject *self, PyObject *value, void *v)
 {
      Body *body = (Body*) self;
-     char *name = MyString_AsString(value);
-     if (!name) return -1;
+     char *name = PyUnicode_AsUTF8(value);
+     if (!name)
+          return -1;
      strncpy(body->obj.o_name, name, MAXNM);
-     free(name);
      body->obj.o_name[MAXNM - 1] = '\0';
      Py_XDECREF(body->name);
      Py_INCREF(value);
@@ -1948,6 +1987,23 @@ static PyMemberDef EarthSatellite_members[] = {
 static char body_doc[] =
      "A celestial body, that can compute() its sky position";
 
+static char earth_satellite_doc[] = "\
+A satellite in orbit around the Earth, usually built by passing the\
+ text of a TLE entry to the `ephem.readtle()` routine. You can read\
+ and write its orbital parameters through the following attributes:\n\
+\n\
+_ap -- argument of perigee at epoch (degrees)\n\
+_decay -- orbit decay rate (revolutions per day-squared)\n\
+_drag -- object drag coefficient (per earth radius)\n\
+_e -- eccentricity\n\
+_epoch -- reference epoch (mjd)\n\
+_inc -- inclination (degrees)\n\
+_M -- mean anomaly (degrees from perigee at epoch)\n\
+_n -- mean motion (revolutions per day)\n\
+_orbit -- integer orbit number of epoch\n\
+_raan -- right ascension of ascending node (degrees)\n\
+";
+
 static PyTypeObject BodyType = {
      PyVarObject_HEAD_INIT(NULL, 0)
      "ephem.Body",
@@ -2431,7 +2487,7 @@ static PyTypeObject EarthSatelliteType = {
      0,				/* tp_setattro */
      0,				/* tp_as_buffer */
      Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
-     body_doc,			/* tp_doc */
+     earth_satellite_doc,	/* tp_doc */
      0,				/* tp_traverse */
      0,				/* tp_clear */
      0,				/* tp_richcompare */
@@ -2510,12 +2566,16 @@ static int separation_arg(PyObject *arg, double *lngi, double *lati)
 static PyObject* separation(PyObject *self, PyObject *args)
 {
      double plat, plng, qlat, qlng;
-     double spy, cpy, px, qx, sqy, cqy;
+     double spy, cpy, px, qx, sqy, cqy, cosine;
      PyObject *p, *q;
      if (!PyArg_ParseTuple(args, "OO:separation", &p, &q)) return 0;
      if (separation_arg(p, &plng, &plat)) return 0;
      if (separation_arg(q, &qlng, &qlat)) return 0;
 
+     /* rounding errors in the trigonometry might return >0 for this case */
+     if ((plat == qlat) && (plng == qlng))
+          return new_Angle(0.0, raddeg(1));
+
      spy = sin (plat);
      cpy = cos (plat);
      px = plng;
@@ -2523,7 +2583,11 @@ static PyObject* separation(PyObject *self, PyObject *args)
      sqy = sin (qlat);
      cqy = cos (qlat);
 
-     return new_Angle(acos(spy*sqy + cpy*cqy*cos(px-qx)), raddeg(1));
+     cosine = spy*sqy + cpy*cqy*cos(px-qx);
+     if (cosine >= 1.0)  /* rounding sometimes makes this >1 */
+          return new_Angle(0.0, raddeg(1));
+
+     return new_Angle(acos(cosine), raddeg(1));
 }
 
 /* Read various database formats, with strings as input. */
@@ -2599,15 +2663,14 @@ static PyObject* readtle(PyObject *self, PyObject *args)
      if (!PyArg_ParseTuple(args, "O!ss:readtle",
 			   &PyUnicode_Type, &name, &l1, &l2))
 	  return 0;
-     l0 = MyString_AsString(name);
-     if (!l0) return 0;
+     l0 = PyUnicode_AsUTF8(name);
+     if (!l0)
+          return 0;
      if (db_tle(l0, l1, l2, &obj)) {
 	  PyErr_SetString(PyExc_ValueError,
 			  "line does not conform to tle format");
-          free(l0);
 	  return 0;
      }
-     free(l0);
      stripped_name = PyObject_CallMethod(name, "strip", 0);
      if (!stripped_name)
 	  return 0;
@@ -2721,7 +2784,7 @@ static PyObject* constellation
 			       "two numeric coordinates");
 	       goto fail;
 	  }
-	  if (!PySequence_Length(position_arg) == 2) {
+	  if (PySequence_Length(position_arg) != 2) {
 	       PyErr_SetString(PyExc_ValueError, "the sequence specifying a "
 			       "position must have exactly two coordinates");
 	       goto fail;
@@ -2977,6 +3040,11 @@ static PyMethodDef libastro_methods[] = {
      {NULL}
 };
 
+#if PY_MAJOR_VERSION == 3
+
+/* PyDoc_STRVAR(doc_module, */
+/* "(Put documentation here.)"); */
+
 static struct PyModuleDef libastro_module = {
      PyModuleDef_HEAD_INIT,
      "_libastro",
@@ -2985,8 +3053,9 @@ static struct PyModuleDef libastro_module = {
      libastro_methods
 };
 
-PyMODINIT_FUNC
-PyInit__libastro(void)
+#endif
+
+PyObject *PyInit__libastro(void)
 {
      PyObject *module;
 
@@ -3023,7 +3092,12 @@ PyInit__libastro(void)
      PyType_Ready(&ParabolicBodyType);
      PyType_Ready(&EarthSatelliteType);
 
+#if PY_MAJOR_VERSION == 2
+     module = Py_InitModule3("_libastro", libastro_methods,
+                             "Astronomical calculations for Python");
+#else
      module = PyModule_Create(&libastro_module);
+#endif
      if (!module) return 0;
 
      {
@@ -3074,5 +3148,20 @@ PyInit__libastro(void)
      /* Tell libastro that we do not have data files anywhere. */
 
      setMoonDir(NULL);
+
+     /* All done! */
+
      return module;
 }
+
+#if PY_MAJOR_VERSION == 2
+#ifdef PyMODINIT_FUNC
+PyMODINIT_FUNC
+#else
+DL_EXPORT(void)
+#endif
+init_libastro(void)
+{
+     PyInit__libastro();
+}
+#endif
