@@ -7,8 +7,9 @@
  *   See http://www.celestrak.com/NORAD/documentation/spacetrk.pdf.
  *
  * A few topocentric routines are also used from the 'orbit' program which is
- *   Copyright (c) 1986,1987,1988,1989,1990 Robert W. Berger N3EMO
- *
+ *   Copyright (c) 1986,1987,1988,1989,1990 Robert W. Berger N3EMO, who has
+ *   granted permission for it to be licensed under the same terms as those
+ *   of the PyEphem package in which this source file is included.
  */
 
 /* define this to use orbit's propagator
@@ -244,13 +245,10 @@ obj_earthsat (Now *np, Obj *op)
 	    ra = op->s_gaera;
 	    dec = op->s_gaedec;
 	}
-	if (epoch != EOD && mjd != epoch)
-	    precess (mjd, epoch, &ra, &dec);
 	op->s_ra = ra;
 	op->s_dec = dec;
-
-	/* Since nutation and aberration not relevant for a satellite,
-	   we store the same RA and dec in its "astrometric" fields */
+	if (epoch != EOD && mjd != epoch)
+	    precess (mjd, epoch, &ra, &dec);
 	op->s_astrora = ra;
 	op->s_astrodec = dec;
 
@@ -401,8 +399,8 @@ double *SatVX, double *SatVY, double *SatVZ)
 static int
 crazyOp (Now *np, Obj *op)
 {
-	/* figure its daft if long enough to decay 1 rev/day */
-	return (fabs(op->es_epoch - mjd) > 1/fabs(op->es_decay));
+	/* toss if more than a year old */
+	return (fabs(op->es_epoch - mjd) > 365);
 }
 
 /* grab the xephem stuff from op and copy into orbit's globals.
@@ -788,4 +786,4 @@ InitOrbitRoutines(double EpochDay, int AtEod)
 }
 
 /* For RCS Only -- Do Not Edit */
-static char *rcsid[2] = {(char *)rcsid, "@(#) $RCSfile: earthsat.c,v $ $Date: 2007/05/09 03:19:29 $ $Revision: 1.10 $ $Name:  $"};
+static char *rcsid[2] = {(char *)rcsid, "@(#) $RCSfile: earthsat.c,v $ $Date: 2009/07/03 23:14:02 $ $Revision: 1.12 $ $Name:  $"};
